from __future__ import annotations

import asyncio
import contextlib
import os
from logging import getLogger
from pathlib import Path
from typing import TYPE_CHECKING, TypeVar

import aioshutil
from aiofiles import ospath
from aiofiles.os import rename, scandir
from typing_extensions import override

from crawlee.base_storage_client import BaseStorageClient
from crawlee.configuration import Configuration
from crawlee.memory_storage_client.dataset_client import DatasetClient
from crawlee.memory_storage_client.dataset_collection_client import DatasetCollectionClient
from crawlee.memory_storage_client.key_value_store_client import KeyValueStoreClient
from crawlee.memory_storage_client.key_value_store_collection_client import KeyValueStoreCollectionClient
from crawlee.memory_storage_client.request_queue_client import RequestQueueClient
from crawlee.memory_storage_client.request_queue_collection_client import RequestQueueCollectionClient

if TYPE_CHECKING:
    from crawlee.base_storage_client.types import ResourceClient

TResourceClient = TypeVar('TResourceClient', DatasetClient, KeyValueStoreClient, RequestQueueClient)

logger = getLogger(__name__)


class MemoryStorageClient(BaseStorageClient):
    """Represents an in-memory storage client for managing datasets, key-value stores, and request queues.

    It emulates in-memory storage similar to the Apify platform, supporting both in-memory and local file system-based
    persistence.

    The behavior of the storage, such as data persistence and metadata writing, can be customized via initialization
    parameters or environment variables.
    """

    _MIGRATING_KEY_VALUE_STORE_DIR_NAME = '__CRAWLEE_MIGRATING_KEY_VALUE_STORE'
    """Name of the directory used to temporarily store files during the migration of the default key-value store."""

    _TEMPORARY_DIR_NAME = '__CRAWLEE_TEMPORARY'
    """Name of the directory used to temporarily store files during purges."""

    def __init__(self, configuration: Configuration | None = None) -> None:
        """Create a new instance.

        Args:
            configuration: Configuration object to use. If None, a default Configuration object will be created.
        """
        self._explicit_configuration = configuration

        self.datasets_handled: list[DatasetClient] = []
        self.key_value_stores_handled: list[KeyValueStoreClient] = []
        self.request_queues_handled: list[RequestQueueClient] = []

        self._purged_on_start = False  # Indicates whether a purge was already performed on this instance.
        self._purge_lock = asyncio.Lock()

    @property
    def _configuration(self) -> Configuration:
        return self._explicit_configuration or Configuration.get_global_configuration()

    @property
    def write_metadata(self) -> bool:
        """Whether to write metadata to the storage."""
        return self._configuration.write_metadata

    @property
    def persist_storage(self) -> bool:
        """Whether to persist the storage."""
        return self._configuration.persist_storage

    @property
    def storage_dir(self) -> str:
        """Path to the storage directory."""
        return self._configuration.storage_dir

    @property
    def datasets_directory(self) -> str:
        """Path to the directory containing datasets."""
        return os.path.join(self.storage_dir, 'datasets')

    @property
    def key_value_stores_directory(self) -> str:
        """Path to the directory containing key-value stores."""
        return os.path.join(self.storage_dir, 'key_value_stores')

    @property
    def request_queues_directory(self) -> str:
        """Path to the directory containing request queues."""
        return os.path.join(self.storage_dir, 'request_queues')

    @override
    def dataset(self, id: str) -> DatasetClient:
        return DatasetClient(
            memory_storage_client=self,
            id=id,
        )

    @override
    def datasets(self) -> DatasetCollectionClient:
        return DatasetCollectionClient(
            memory_storage_client=self,
        )

    @override
    def key_value_store(self, id: str) -> KeyValueStoreClient:
        return KeyValueStoreClient(
            memory_storage_client=self,
            id=id,
        )

    @override
    def key_value_stores(self) -> KeyValueStoreCollectionClient:
        return KeyValueStoreCollectionClient(
            memory_storage_client=self,
        )

    @override
    def request_queue(self, id: str) -> RequestQueueClient:
        return RequestQueueClient(
            memory_storage_client=self,
            id=id,
        )

    @override
    def request_queues(self) -> RequestQueueCollectionClient:
        return RequestQueueCollectionClient(
            memory_storage_client=self,
        )

    @override
    async def purge_on_start(self) -> None:
        # Optimistic, non-blocking check
        if self._purged_on_start is True:
            logger.debug('Storage was already purged on start.')
            return

        async with self._purge_lock:
            # Another check under the lock just to be sure
            if self._purged_on_start is True:
                # Mypy doesn't understand that the _purged_on_start can change while we're getting the async lock
                return  # type: ignore[unreachable]

            await self._purge_default_storages()
            self._purged_on_start = True

    def get_cached_resource_client(
        self, resource_client_class: type[TResourceClient], id: str | None, name: str | None
    ) -> TResourceClient | None:
        """Try to return a resource client from the internal cache."""
        if issubclass(resource_client_class, DatasetClient):
            cache = self.datasets_handled
        elif issubclass(resource_client_class, KeyValueStoreClient):
            cache = self.key_value_stores_handled
        elif issubclass(resource_client_class, RequestQueueClient):
            cache = self.request_queues_handled
        else:
            return None

        for storage_client in cache:
            if storage_client.id == id or (
                storage_client.name and name and storage_client.name.lower() == name.lower()
            ):
                return storage_client  # pyright: ignore

        return None

    def add_resource_client_to_cache(self, resource_client: ResourceClient) -> None:
        """Add a new resource client to the internal cache."""
        if isinstance(resource_client, DatasetClient):
            self.datasets_handled.append(resource_client)
        if isinstance(resource_client, KeyValueStoreClient):
            self.key_value_stores_handled.append(resource_client)
        if isinstance(resource_client, RequestQueueClient):
            self.request_queues_handled.append(resource_client)

    async def _purge_default_storages(self) -> None:
        """Cleans up the storage directories, preparing the environment for a new run.

        It aims to remove residues from previous executions to avoid data contamination between runs.

        It specifically targets:
         - The local directory containing the default dataset.
         - All records from the default key-value store in the local directory, except for the 'INPUT' key.
         - The local directory containing the default request queue.
        """
        # Key-value stores
        if await ospath.exists(self.key_value_stores_directory):
            key_value_store_folders = await scandir(self.key_value_stores_directory)
            for key_value_store_folder in key_value_store_folders:
                if key_value_store_folder.name.startswith(
                    self._TEMPORARY_DIR_NAME
                ) or key_value_store_folder.name.startswith('__OLD'):
                    await self._batch_remove_files(key_value_store_folder.path)
                elif key_value_store_folder.name == self._configuration.default_key_value_store_id:
                    await self._handle_default_key_value_store(key_value_store_folder.path)

        # Datasets
        if await ospath.exists(self.datasets_directory):
            dataset_folders = await scandir(self.datasets_directory)
            for dataset_folder in dataset_folders:
                if dataset_folder.name == self._configuration.default_dataset_id or dataset_folder.name.startswith(
                    self._TEMPORARY_DIR_NAME
                ):
                    await self._batch_remove_files(dataset_folder.path)

        # Request queues
        if await ospath.exists(self.request_queues_directory):
            request_queue_folders = await scandir(self.request_queues_directory)
            for request_queue_folder in request_queue_folders:
                if (
                    request_queue_folder.name == self._configuration.default_request_queue_id
                    or request_queue_folder.name.startswith(self._TEMPORARY_DIR_NAME)
                ):
                    await self._batch_remove_files(request_queue_folder.path)

    async def _handle_default_key_value_store(self, folder: str) -> None:
        """Manages the cleanup of the default key-value store.

        It removes all files to ensure a clean state except for a set of predefined input keys (`possible_input_keys`).

        Args:
            folder: Path to the default key-value store directory to clean.
        """
        folder_exists = await ospath.exists(folder)
        temporary_path = os.path.normpath(os.path.join(folder, '..', self._MIGRATING_KEY_VALUE_STORE_DIR_NAME))

        # For optimization, we want to only attempt to copy a few files from the default key-value store
        possible_input_keys = [
            'INPUT',
            'INPUT.json',
            'INPUT.bin',
            'INPUT.txt',
        ]

        if folder_exists:
            # Create a temporary folder to save important files in
            Path(temporary_path).mkdir(parents=True, exist_ok=True)

            # Go through each file and save the ones that are important
            for entity in possible_input_keys:
                original_file_path = os.path.join(folder, entity)
                temp_file_path = os.path.join(temporary_path, entity)
                with contextlib.suppress(Exception):
                    await rename(original_file_path, temp_file_path)

            # Remove the original folder and all its content
            counter = 0
            temp_path_for_old_folder = os.path.normpath(os.path.join(folder, f'../__OLD_DEFAULT_{counter}__'))
            done = False
            try:
                while not done:
                    await rename(folder, temp_path_for_old_folder)
                    done = True
            except Exception:
                counter += 1
                temp_path_for_old_folder = os.path.normpath(os.path.join(folder, f'../__OLD_DEFAULT_{counter}__'))

            # Replace the temporary folder with the original folder
            await rename(temporary_path, folder)

            # Remove the old folder
            await self._batch_remove_files(temp_path_for_old_folder)

    async def _batch_remove_files(self, folder: str, counter: int = 0) -> None:
        """Removes a folder and its contents in batches to minimize blocking time.

        This method first renames the target folder to a temporary name, then deletes the temporary folder,
        allowing the file system operations to proceed without hindering other asynchronous tasks.

        Args:
            folder: The directory path to remove.
            counter: A counter used for generating temporary directory names in case of conflicts.
        """
        folder_exists = await ospath.exists(folder)

        if folder_exists:
            temporary_folder = (
                folder
                if os.path.basename(folder).startswith(f'{self._TEMPORARY_DIR_NAME}_')
                else os.path.normpath(os.path.join(folder, '..', f'{self._TEMPORARY_DIR_NAME}_{counter}'))
            )

            try:
                # Rename the old folder to the new one to allow background deletions
                await rename(folder, temporary_folder)
            except Exception:
                # Folder exists already, try again with an incremented counter
                return await self._batch_remove_files(folder, counter + 1)

            await aioshutil.rmtree(temporary_folder, ignore_errors=True)
        return None



# Inspiration: https://github.com/apify/crawlee/blob/v3.9.2/packages/core/src/crawlers/statistics.ts
from __future__ import annotations

import math
from datetime import datetime, timedelta, timezone
from logging import getLogger
from typing import TYPE_CHECKING, Any, Generic, cast

from typing_extensions import Self, TypeVar

from crawlee._utils.recurring_task import RecurringTask
from crawlee.events import LocalEventManager
from crawlee.events.types import Event, EventPersistStateData
from crawlee.statistics import FinalStatistics, StatisticsPersistedState, StatisticsState
from crawlee.statistics.error_tracker import ErrorTracker
from crawlee.storages import KeyValueStore

if TYPE_CHECKING:
    from types import TracebackType

    from crawlee.events import EventManager

TStatisticsState = TypeVar('TStatisticsState', bound=StatisticsState, default=StatisticsState)

logger = getLogger(__name__)


class RequestProcessingRecord:
    """Tracks information about the processing of a request."""

    def __init__(self) -> None:
        self._last_run_at: datetime | None = None
        self._runs = 0
        self.duration: timedelta | None = None

    def run(self) -> int:
        """Mark the job as started."""
        self._last_run_at = datetime.now(timezone.utc)
        self._runs += 1
        return self._runs

    def finish(self) -> timedelta:
        """Mark the job as finished."""
        if self._last_run_at is None:
            raise RuntimeError('Invalid state')

        self.duration = datetime.now(timezone.utc) - self._last_run_at
        return self.duration

    @property
    def retry_count(self) -> int:
        """Number of times the job has been retried."""
        return max(0, self._runs - 1)


class Statistics(Generic[TStatisticsState]):
    """An interface to collecting and logging runtime statistics for requests.

    All information is saved to the key value store so that it persists between migrations, abortions and resurrections.
    """

    __next_id = 0

    def __init__(
        self,
        *,
        event_manager: EventManager | None = None,
        persistence_enabled: bool = False,
        persist_state_kvs_name: str = 'default',
        persist_state_key: str | None = None,
        key_value_store: KeyValueStore | None = None,
        log_message: str = 'Statistics',
        log_interval: timedelta = timedelta(minutes=1),
        state_model: type[TStatisticsState] = cast(Any, StatisticsState),  # noqa: B008 - in an ideal world, TStatisticsState would be inferred from this argument, but I haven't managed to do that
    ) -> None:
        self._id = Statistics.__next_id
        Statistics.__next_id += 1

        self._state_model = state_model
        self.state: StatisticsState = self._state_model()
        self._instance_start: datetime | None = None
        self._retry_histogram = dict[int, int]()

        self.error_tracker = ErrorTracker()
        self.error_tracker_retry = ErrorTracker()

        self._events = event_manager or LocalEventManager()

        self._requests_in_progress = dict[str, RequestProcessingRecord]()

        if persist_state_key is None:
            persist_state_key = f'SDK_CRAWLER_STATISTICS_{self._id}'

        self._persistence_enabled = persistence_enabled
        self._persist_state_key = persist_state_key
        self._persist_state_kvs_name = persist_state_kvs_name
        self._key_value_store: KeyValueStore | None = key_value_store

        self._log_message = log_message
        self._periodic_logger = RecurringTask(self._log, log_interval)

    async def __aenter__(self) -> Self:
        """Subscribe to events and start collecting statistics."""
        self._instance_start = datetime.now(timezone.utc)

        if self.state.crawler_started_at is None:
            self.state.crawler_started_at = datetime.now(timezone.utc)

        if self._key_value_store is None:
            self._key_value_store = await KeyValueStore.open(name=self._persist_state_kvs_name)

        await self._maybe_load_statistics()
        self._events.on(event=Event.PERSIST_STATE, listener=self._persist_state)

        self._periodic_logger.start()

        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        exc_traceback: TracebackType | None,
    ) -> None:
        """Stop collecting statistics."""
        self.state.crawler_finished_at = datetime.now(timezone.utc)
        self._events.off(event=Event.PERSIST_STATE, listener=self._persist_state)
        await self._periodic_logger.stop()
        await self._persist_state(event_data=EventPersistStateData(is_migrating=False))

    def register_status_code(self, code: int) -> None:
        """Increment the number of times a status code has been received."""
        self.state.requests_with_status_code.setdefault(str(code), 0)
        self.state.requests_with_status_code[str(code)] += 1

    def record_request_processing_start(self, request_id_or_key: str) -> None:
        """Mark a request as started."""
        record = self._requests_in_progress.get(request_id_or_key, RequestProcessingRecord())
        record.run()
        self._requests_in_progress[request_id_or_key] = record

    def record_request_processing_finish(self, request_id_or_key: str) -> None:
        """Mark a request as finished."""
        record = self._requests_in_progress.get(request_id_or_key)
        if record is None:
            return

        duration = record.finish()
        self.state.requests_finished += 1
        self.state.request_total_finished_duration += duration
        self._save_retry_count_for_request(record)
        self.state.request_min_duration = min(
            self.state.request_min_duration if self.state.request_min_duration is not None else timedelta.max, duration
        )
        self.state.request_max_duration = min(
            self.state.request_max_duration if self.state.request_max_duration is not None else timedelta(), duration
        )

        del self._requests_in_progress[request_id_or_key]

    def record_request_processing_failure(self, request_id_or_key: str) -> None:
        """Mark a request as failed."""
        record = self._requests_in_progress.get(request_id_or_key)
        if record is None:
            return

        self.state.request_total_failed_duration += record.finish()
        self.state.requests_failed += 1
        self._save_retry_count_for_request(record)

        del self._requests_in_progress[request_id_or_key]

    def calculate(self) -> FinalStatistics:
        """Calculate the current statistics."""
        if self._instance_start is None:
            raise RuntimeError('The Statistics object is not initialized')

        crawler_runtime = datetime.now(timezone.utc) - self._instance_start
        total_minutes = crawler_runtime.total_seconds() / 60

        return FinalStatistics(
            request_avg_failed_duration=(self.state.request_total_failed_duration / self.state.requests_failed)
            if self.state.requests_failed
            else None,
            request_avg_finished_duration=(self.state.request_total_finished_duration / self.state.requests_finished)
            if self.state.requests_finished
            else None,
            requests_finished_per_minute=round(self.state.requests_finished / total_minutes) if total_minutes else 0,
            requests_failed_per_minute=math.floor(self.state.requests_failed / total_minutes) if total_minutes else 0,
            request_total_duration=self.state.request_total_finished_duration
            + self.state.request_total_failed_duration,
            requests_total=self.state.requests_failed + self.state.requests_finished,
            crawler_runtime=crawler_runtime,
            requests_finished=self.state.requests_finished,
            requests_failed=self.state.requests_failed,
            retry_histogram=[
                self._retry_histogram.get(retry_count, 0)
                for retry_count in range(max(self._retry_histogram.keys(), default=0) + 1)
            ],
        )

    async def reset(self) -> None:
        """Reset the statistics to their defaults and remove any persistent state."""
        self.state = self._state_model()
        self.error_tracker = ErrorTracker()
        self.error_tracker_retry = ErrorTracker()
        self._retry_histogram.clear()
        self._requests_in_progress.clear()

        if self._persistence_enabled and self._key_value_store:
            await self._key_value_store.set_value(self._persist_state_key, None)

    def _log(self) -> None:
        stats = self.calculate()
        logger.info(f'{self._log_message} {stats}')

    async def _maybe_load_statistics(self) -> None:
        if not self._persistence_enabled:
            return

        if not self._key_value_store:
            return

        stored_state = await self._key_value_store.get_value(self._persist_state_key, cast(Any, {}))

        saved_state = self.state.__class__.model_validate(stored_state)
        self.state = saved_state

        if saved_state.stats_persisted_at is not None and saved_state.crawler_last_started_at:
            self._instance_start = datetime.now(timezone.utc) - (
                saved_state.stats_persisted_at - saved_state.crawler_last_started_at
            )
        elif saved_state.crawler_last_started_at:
            self._instance_start = saved_state.crawler_last_started_at

    async def _persist_state(self, event_data: EventPersistStateData) -> None:
        logger.debug(f'Persisting state of the Statistics (event_data={event_data}).')

        if not self._persistence_enabled:
            return

        if not self._key_value_store:
            return

        if not self._instance_start:
            return

        final_statistics = self.calculate()
        persisted_state = StatisticsPersistedState(
            stats_id=self._id,
            stats_persisted_at=datetime.now(timezone.utc),
            crawler_last_started_at=self._instance_start,
            request_total_duration=final_statistics.request_total_duration,
            request_avg_failed_duration=final_statistics.request_avg_failed_duration,
            request_avg_finished_duration=final_statistics.request_avg_finished_duration,
            requests_total=final_statistics.requests_total,
            request_retry_histogram=final_statistics.retry_histogram,
        )

        logger.debug('Persisting state')

        await self._key_value_store.set_value(
            self._persist_state_key,
            self.state.model_dump(mode='json', by_alias=True) | persisted_state.model_dump(mode='json', by_alias=True),
            'application/json',
        )

    def _save_retry_count_for_request(self, record: RequestProcessingRecord) -> None:
        retry_count = record.retry_count

        if retry_count:
            self.state.requests_retries += 1

        self._retry_histogram.setdefault(retry_count, 0)
        self._retry_histogram[retry_count] += 1



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from crawlee.models import KeyValueStoreListPage, KeyValueStoreMetadata


class BaseKeyValueStoreCollectionClient(ABC):
    """Abstract base class for key-value store collection clients.

    This collection client handles operations that involve multiple instances of a given resource type.
    """

    @abstractmethod
    async def get_or_create(
        self,
        *,
        id: str | None = None,
        name: str | None = None,
        schema: dict | None = None,
    ) -> KeyValueStoreMetadata:
        """Retrieve an existing key-value store by its name or ID, or create a new one if it does not exist.

        Args:
            id: Optional ID of the key-value store to retrieve or create. If provided, the method will attempt
                to find a key-value store with the ID.

            name: Optional name of the key-value store resource to retrieve or create. If provided, the method will
                attempt to find a key-value store with this name.

            schema: Optional schema for the key-value store resource to be created.

        Returns:
            Metadata object containing the information of the retrieved or created key-value store.
        """

    @abstractmethod
    async def list(
        self,
        *,
        unnamed: bool = False,
        limit: int | None = None,
        offset: int | None = None,
        desc: bool = False,
    ) -> KeyValueStoreListPage:
        """List the available key-value stores.

        Args:
            unnamed: Whether to list only the unnamed key-value stores.
            limit: Maximum number of key-value stores to return.
            offset: Number of key-value stores to skip from the beginning of the list.
            desc: Whether to sort the key-value stores in descending order.

        Returns:
            The list of available key-value stores matching the specified filters.
        """



# ruff: noqa: TCH002 TCH003

from __future__ import annotations

from datetime import datetime, timedelta
from typing import Annotated, Any

from dateutil import parser
from pydantic import BaseModel, ConfigDict, Field, field_validator


class SessionModel(BaseModel):
    """Model for a Session object."""

    model_config = ConfigDict(populate_by_name=True)

    id: Annotated[str, Field(alias='id')]
    max_age: Annotated[timedelta, Field(alias='maxAge')]
    user_data: Annotated[dict, Field(alias='userData')]
    max_error_score: Annotated[float, Field(alias='maxErrorScore')]
    error_score_decrement: Annotated[float, Field(alias='errorScoreDecrement')]
    created_at: Annotated[datetime, Field(alias='createdAt')]
    usage_count: Annotated[int, Field(alias='usageCount')]
    max_usage_count: Annotated[int, Field(alias='maxUsageCount')]
    error_score: Annotated[float, Field(alias='errorScore')]
    cookies: Annotated[dict, Field(alias='cookies')]
    blocked_status_codes: Annotated[list[int], Field(alias='blockedStatusCodes')]

    @field_validator('max_age', mode='before')
    @classmethod
    def parse_max_age(cls, value: Any) -> timedelta:
        """Try to parse max_age field into a timedelta object."""
        if isinstance(value, timedelta):
            return value

        if isinstance(value, str):
            try:
                parsed_time = parser.parse(value)
                return timedelta(hours=parsed_time.hour, minutes=parsed_time.minute, seconds=parsed_time.second)
            except ValueError as exc:
                raise ValueError(f"Invalid time format for max_age. Expected 'HH:MM:SS', got {value}") from exc

        raise ValueError('Invalid data type for max_age')


class SessionPoolModel(BaseModel):
    """Model for a SessionPool object."""

    model_config = ConfigDict(populate_by_name=True)

    persistence_enabled: Annotated[bool, Field(alias='persistenceEnabled')]
    persist_state_kvs_name: Annotated[str, Field(alias='persistStateKvsName')]
    persist_state_key: Annotated[str, Field(alias='persistStateKey')]
    max_pool_size: Annotated[int, Field(alias='maxPoolSize')]
    session_count: Annotated[int, Field(alias='sessionCount')]
    usable_session_count: Annotated[int, Field(alias='usableSessionCount')]
    retired_session_count: Annotated[int, Field(alias='retiredSessionCount')]
    sessions: Annotated[list[SessionModel], Field(alias='sessions')]



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Sequence

    from crawlee.models import (
        BatchRequestsOperationResponse,
        ProcessedRequest,
        ProlongRequestLockResponse,
        Request,
        RequestListResponse,
        RequestQueueHead,
        RequestQueueHeadWithLocks,
        RequestQueueMetadata,
    )


class BaseRequestQueueClient(ABC):
    """Abstract base class for request queue resource clients.

    These clients are specific to the type of resource they manage and operate under a designated storage
    client, like a memory storage client.
    """

    @abstractmethod
    async def get(self) -> RequestQueueMetadata | None:
        """Get metadata about the request queue being managed by this client.

        Returns:
            An object containing the request queue's details, or None if the request queue does not exist.
        """

    @abstractmethod
    async def update(
        self,
        *,
        name: str | None = None,
    ) -> RequestQueueMetadata:
        """Update the request queue metadata.

        Args:
            name: New new name for the request queue.

        Returns:
            An object reflecting the updated request queue metadata.
        """

    @abstractmethod
    async def delete(self) -> None:
        """Permanently delete the request queue managed by this client."""

    @abstractmethod
    async def list_head(self, *, limit: int | None = None) -> RequestQueueHead:
        """Retrieve a given number of requests from the beginning of the queue.

        Args:
            limit: How many requests to retrieve

        Returns:
            The desired number of requests from the beginning of the queue.
        """

    @abstractmethod
    async def list_and_lock_head(self, *, lock_secs: int, limit: int | None = None) -> RequestQueueHeadWithLocks:
        """Fetch and lock a specified number of requests from the start of the queue.

        Retrieves and locks the first few requests of a queue for the specified duration. This prevents the requests
        from being fetched by another client until the lock expires.

        Args:
            lock_secs: Duration for which the requests are locked, in seconds.
            limit: Maximum number of requests to retrieve and lock.

        Returns:
            The desired number of locked requests from the beginning of the queue.
        """

    @abstractmethod
    async def add_request(
        self,
        request: Request,
        *,
        forefront: bool = False,
    ) -> ProcessedRequest:
        """Add a request to the queue.

        Args:
            request: The request to add to the queue
            forefront: Whether to add the request to the head or the end of the queue

        Returns:
            Request queue operation information.
        """

    @abstractmethod
    async def batch_add_requests(
        self,
        requests: Sequence[Request],
        *,
        forefront: bool = False,
    ) -> BatchRequestsOperationResponse:
        """Add a batch of requests to the queue.

        Args:
            requests: The requests to add to the queue
            forefront: Whether to add the requests to the head or the end of the queue

        Returns:
            Request queue batch operation information.
        """

    @abstractmethod
    async def get_request(self, request_id: str) -> Request | None:
        """Retrieve a request from the queue.

        Args:
            request_id: ID of the request to retrieve

        Returns:
            The retrieved request, or None, if it did not exist.
        """

    @abstractmethod
    async def update_request(
        self,
        request: Request,
        *,
        forefront: bool = False,
    ) -> ProcessedRequest:
        """Update a request in the queue.

        Args:
            request: The updated request
            forefront: Whether to put the updated request in the beginning or the end of the queue

        Returns:
            The updated request
        """

    @abstractmethod
    async def delete_request(self, request_id: str) -> None:
        """Delete a request from the queue.

        Args:
            request_id: ID of the request to delete.
        """

    @abstractmethod
    async def prolong_request_lock(
        self,
        request_id: str,
        *,
        forefront: bool = False,
        lock_secs: int,
    ) -> ProlongRequestLockResponse:
        """Prolong the lock on a specific request in the queue.

        Args:
            request_id: The identifier of the request whose lock is to be prolonged.
            forefront: Whether to put the request in the beginning or the end of the queue after lock expires.
            lock_secs: The additional amount of time, in seconds, that the request will remain locked.
        """

    @abstractmethod
    async def delete_request_lock(
        self,
        request_id: str,
        *,
        forefront: bool = False,
    ) -> None:
        """Delete the lock on a specific request in the queue.

        Args:
            request_id: ID of the request to delete the lock
            forefront: Whether to put the request in the beginning or the end of the queue after the lock is deleted.
        """

    @abstractmethod
    async def batch_delete_requests(self, requests: list[Request]) -> BatchRequestsOperationResponse:
        """Delete given requests from the queue.

        Args:
            requests: The requests to delete from the queue.
        """

    @abstractmethod
    async def list_requests(
        self,
        *,
        limit: int | None = None,
        exclusive_start_id: str | None = None,
    ) -> RequestListResponse:
        """List requests from the queue.

        Args:
            limit: How many requests to retrieve.
            exclusive_start_id: All requests up to this one (including) are skipped from the result.
        """



from __future__ import annotations

from enum import Enum
from typing import Any, Union

# Type for representing json-serializable values. It's close enough to the real thing supported
# by json.parse, and the best we can do until mypy supports recursive types. It was suggested
# in a discussion with (and approved by) Guido van Rossum, so I'd consider it correct enough.
JSONSerializable = Union[str, int, float, bool, None, dict[str, Any], list[Any]]


class StorageTypes(str, Enum):
    """Possible Crawlee storage types."""

    DATASET = 'Dataset'
    KEY_VALUE_STORE = 'Key-value store'
    REQUEST_QUEUE = 'Request queue'



try:
    from .playwright_crawler import PlaywrightCrawler
    from .types import PlaywrightCrawlingContext
except ImportError as exc:
    raise ImportError(
        'To import anything from this subpacakge, you need to install the "playwright" extra. '
        'For example, if you use pip, run "pip install crawlee[playwright]".',
    ) from exc

__all__ = ['PlaywrightCrawler', 'PlaywrightCrawlingContext']



# ruff: noqa: TCH003 TCH002 TCH001

from __future__ import annotations

from datetime import timedelta
from typing import Annotated, ClassVar, cast

from pydantic import AliasChoices, Field
from pydantic_settings import BaseSettings, SettingsConfigDict
from typing_extensions import Self

from crawlee._utils.models import timedelta_ms


class Configuration(BaseSettings):
    """Configuration of the Crawler.

    Args:
        internal_timeout: timeout for internal operations such as marking a request as processed
        verbose_log: allows verbose logging
        default_storage_id: The default storage ID.
        purge_on_start: Whether to purge the storage on start.
    """

    _default_instance: ClassVar[Self | None] = None

    model_config = SettingsConfigDict(populate_by_name=True)

    internal_timeout: Annotated[timedelta | None, Field(alias='crawlee_internal_timeout')] = None

    verbose_log: Annotated[bool, Field(alias='crawlee_verbose_log')] = False

    default_browser_path: Annotated[
        str | None,
        Field(
            validation_alias=AliasChoices(
                'apify_default_browser_path',
                'crawlee_default_browser_path',
            )
        ),
    ] = None

    disable_browser_sandbox: Annotated[
        bool,
        Field(
            validation_alias=AliasChoices(
                'apify_disable_browser_sandbox',
                'crawlee_disable_browser_sandbox',
            )
        ),
    ] = False

    log_level: Annotated[
        int,
        Field(
            validation_alias=AliasChoices(
                'apify_log_level',
                'crawlee_log_level',
            )
        ),
    ] = 4  # INFO

    default_dataset_id: Annotated[
        str,
        Field(
            validation_alias=AliasChoices(
                'actor_default_dataset_id',
                'apify_default_dataset_id',
                'crawlee_default_dataset_id',
            )
        ),
    ] = 'default'

    default_key_value_store_id: Annotated[
        str,
        Field(
            validation_alias=AliasChoices(
                'actor_default_key_value_store_id',
                'apify_default_key_value_store_id',
                'crawlee_default_key_value_store_id',
            )
        ),
    ] = 'default'

    default_request_queue_id: Annotated[
        str,
        Field(
            validation_alias=AliasChoices(
                'actor_default_request_queue_id',
                'apify_default_request_queue_id',
                'crawlee_default_request_queue_id',
            )
        ),
    ] = 'default'

    purge_on_start: Annotated[
        bool,
        Field(
            validation_alias=AliasChoices(
                'apify_purge_on_start',
                'crawlee_purge_on_start',
            )
        ),
    ] = True

    write_metadata: Annotated[bool, Field(alias='crawlee_write_metadata')] = True

    persist_storage: Annotated[
        bool,
        Field(
            validation_alias=AliasChoices(
                'apify_persist_storage',
                'crawlee_persist_storage',
            )
        ),
    ] = True

    persist_state_interval: Annotated[
        timedelta_ms,
        Field(
            validation_alias=AliasChoices(
                'apify_persist_state_interval_millis',
                'crawlee_persist_state_interval_millis',
            )
        ),
    ] = timedelta(minutes=1)

    system_info_interval: Annotated[
        timedelta_ms,
        Field(
            validation_alias=AliasChoices(
                'apify_system_info_interval_millis',
                'crawlee_system_info_interval_millis',
            )
        ),
    ] = timedelta(seconds=1)

    max_used_cpu_ratio: Annotated[
        float,
        Field(
            validation_alias=AliasChoices(
                'apify_max_used_cpu_ratio',
                'crawlee_max_used_cpu_ratio',
            )
        ),
    ] = 0.95

    memory_mbytes: Annotated[
        int | None,
        Field(
            validation_alias=AliasChoices(
                'actor_memory_mbytes',
                'apify_memory_mbytes',
                'crawlee_memory_mbytes',
            )
        ),
    ] = None

    available_memory_ratio: Annotated[
        float | None,
        Field(
            validation_alias=AliasChoices(
                'apify_available_memory_ratio',
                'crawlee_available_memory_ratio',
            )
        ),
    ] = None

    storage_dir: Annotated[
        str,
        Field(
            validation_alias=AliasChoices(
                'apify_local_storage_dir',
                'crawlee_storage_dir',
            ),
        ),
    ] = './storage'

    chrome_executable_path: Annotated[
        str | None,
        Field(
            validation_alias=AliasChoices(
                'apify_chrome_executable_path',
                'crawlee_chrome_executable_path',
            )
        ),
    ] = None

    headless: Annotated[
        bool,
        Field(
            validation_alias=AliasChoices(
                'apify_headless',
                'crawlee_headless',
            )
        ),
    ] = True

    xvfb: Annotated[
        bool,
        Field(
            validation_alias=AliasChoices(
                'apify_xvfb',
                'crawlee_xvfb',
            )
        ),
    ] = False

    in_cloud: Annotated[bool, Field(alias='crawlee_in_cloud')] = False

    @classmethod
    def get_global_configuration(cls) -> Self:
        """Retrieve the global instance of the configuration."""
        if Configuration._default_instance is None:
            Configuration._default_instance = cls()

        return cast(Self, Configuration._default_instance)



# Inspiration: https://github.com/apify/crawlee/blob/v3.9.0/packages/core/src/session_pool/session_pool.ts

from __future__ import annotations

import random
from logging import getLogger
from typing import TYPE_CHECKING, Callable, Literal, overload

from crawlee.events.types import Event, EventPersistStateData
from crawlee.sessions import Session
from crawlee.sessions.models import SessionPoolModel
from crawlee.storages import KeyValueStore

if TYPE_CHECKING:
    from types import TracebackType

    from crawlee.events import EventManager

logger = getLogger(__name__)

CreateSessionFunctionType = Callable[[], Session]


class SessionPool:
    """Session pool is a pool of sessions that are rotated based on the usage count or age."""

    def __init__(
        self,
        *,
        max_pool_size: int = 1000,
        create_session_settings: dict | None = None,
        create_session_function: CreateSessionFunctionType | None = None,
        event_manager: EventManager | None = None,
        persistence_enabled: bool = False,
        persist_state_kvs_name: str = 'default',
        persist_state_key: str = 'CRAWLEE_SESSION_POOL_STATE',
    ) -> None:
        """Create a new instance.

        Args:
            max_pool_size: Maximum number of sessions to maintain in the pool. You can add more sessions to the pool
                by using the `add_session` method.

            create_session_settings: Settings for creating new session instances. If None, default settings will
                be used. Do not set it if you are providing a `create_session_function`.

            create_session_function: A callable to create new session instances. If None, a default session settings
                will be used. Do not set it if you are providing `create_session_settings`.

            event_manager: The event manager to handle events like persist state.

            persistence_enabled: Flag to enable or disable state persistence of the pool. If it is enabled, make sure
                to provide an event manager to handle the events.

            persist_state_kvs_name: The name of the `KeyValueStore` used for state persistence.

            persist_state_key: The key under which the session pool's state is stored in the `KeyValueStore`.
        """
        self._max_pool_size = max_pool_size
        self._session_settings = create_session_settings or {}
        self._create_session_function = create_session_function
        self._event_manager = event_manager
        self._persistence_enabled = persistence_enabled
        self._persist_state_kvs_name = persist_state_kvs_name
        self._persist_state_key = persist_state_key

        if self._create_session_function and self._session_settings:
            raise ValueError('Both `create_session_settings` and `create_session_function` cannot be provided.')

        if self._persistence_enabled and not self._event_manager:
            raise ValueError('Persistence is enabled, but no event manager was provided.')

        # Internal non-configurable attributes
        self._kvs: KeyValueStore | None = None
        self._sessions: dict[str, Session] = {}

    def __repr__(self) -> str:
        """Get a string representation."""
        return f'<{self.__class__.__name__} {self.get_state(as_dict=False)}>'

    @property
    def session_count(self) -> int:
        """Get the total number of sessions currently maintained in the pool."""
        return len(self._sessions)

    @property
    def usable_session_count(self) -> int:
        """Get the number of sessions that are currently usable."""
        return len([session for _, session in self._sessions.items() if session.is_usable])

    @property
    def retired_session_count(self) -> int:
        """Get the number of sessions that are no longer usable."""
        return self.session_count - self.usable_session_count

    async def __aenter__(self) -> SessionPool:
        """Initialize the pool upon entering the context manager."""
        if self._persistence_enabled and self._event_manager:
            self._kvs = await KeyValueStore.open(name=self._persist_state_kvs_name)

            # Attempt to restore the previously persisted state.
            was_restored = await self._try_to_restore_previous_state()

            # If the pool could not be restored, initialize it with new sessions.
            if not was_restored:
                await self._fill_sessions_to_max()

            # Register an event listener for persisting the session pool state.
            self._event_manager.on(event=Event.PERSIST_STATE, listener=self._persist_state)
        # If persistence is disabled, just fill the pool with sessions.
        else:
            await self._fill_sessions_to_max()

        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        exc_traceback: TracebackType | None,
    ) -> None:
        """Deinitialize the pool upon exiting the context manager."""
        if self._persistence_enabled and self._event_manager:
            # Remove the event listener for state persistence.
            self._event_manager.off(event=Event.PERSIST_STATE, listener=self._persist_state)

            # Persist the final state of the session pool.
            await self._persist_state(event_data=EventPersistStateData(is_migrating=False))

    @overload
    def get_state(self, *, as_dict: Literal[True]) -> dict: ...

    @overload
    def get_state(self, *, as_dict: Literal[False]) -> SessionPoolModel: ...

    def get_state(self, *, as_dict: bool = False) -> SessionPoolModel | dict:
        """Retrieve the current state of the pool either as a model or as a dictionary."""
        model = SessionPoolModel(
            persistence_enabled=self._persistence_enabled,
            persist_state_kvs_name=self._persist_state_kvs_name,
            persist_state_key=self._persist_state_key,
            max_pool_size=self._max_pool_size,
            session_count=self.session_count,
            usable_session_count=self.usable_session_count,
            retired_session_count=self.retired_session_count,
            sessions=[session.get_state(as_dict=False) for _, session in self._sessions.items()],
        )
        if as_dict:
            return model.model_dump()
        return model

    def add_session(self, session: Session) -> None:
        """Add a specific session to the pool.

        This is intened only for the cases when you want to add a session that was created outside of the pool.
        Otherwise, the pool will create new sessions automatically.
        """
        if session.id in self._sessions:
            logger.warning(f'Session with ID {session.id} already exists in the pool.')
            return
        self._sessions[session.id] = session

    async def get_session(self) -> Session:
        """Retrieve a random session from the pool.

        This method first ensures the session pool is at its maximum capacity. If the random session is not usable,
        retired sessions are removed and a new session is created and returned.

        Returns:
            The session object.
        """
        await self._fill_sessions_to_max()
        session = self._get_random_session()

        if session.is_usable:
            return session

        # If the random session is not usable, clean up and create a new session
        self._remove_retired_sessions()
        return await self._create_new_session()

    async def get_session_by_id(self, session_id: str) -> Session | None:
        """Retrieve a session by ID from the pool.

        This method first ensures the session pool is at its maximum capacity. It then tries to retrieve a specific
        session by ID. If the session is not found or not usable, `None` is returned.

        Args:
            session_id: The ID of the session to retrieve.

        Returns:
            The session object if found and usable, otherwise `None`.
        """
        await self._fill_sessions_to_max()
        session = self._sessions.get(session_id)

        if not session:
            logger.warning(f'Session with ID {session_id} not found.')
            return None

        if not session.is_usable:
            logger.warning(f'Session with ID {session_id} is not usable.')
            return None

        return session

    async def reset_store(self) -> None:
        """Reset the KVS where the pool state is persisted."""
        if not self._persistence_enabled:
            logger.debug('Persistence is disabled; skipping the reset of the store.')
            return

        if not self._kvs:
            logger.warning('SessionPool reset failed: KVS not initialized. Did you forget to call __aenter__?')
            return

        await self._kvs.set_value(key=self._persist_state_key, value=None)

    async def _create_new_session(self) -> Session:
        """Create a new session, add it to the pool and return it."""
        if self._create_session_function:
            new_session = self._create_session_function()
        else:
            new_session = Session(**self._session_settings)
        self._sessions[new_session.id] = new_session
        return new_session

    async def _fill_sessions_to_max(self) -> None:
        """Fill the pool with sessions to the maximum size."""
        for _ in range(self._max_pool_size - self.session_count):
            await self._create_new_session()

    def _get_random_session(self) -> Session:
        """Get a random session from the pool."""
        keys = list(self._sessions.keys())
        if not keys:
            raise ValueError('No sessions available in the pool.')
        key = random.choice(keys)
        return self._sessions[key]

    def _remove_retired_sessions(self) -> None:
        """Remove all sessions from the pool that are no longer usable."""
        self._sessions = {session_id: session for session_id, session in self._sessions.items() if session.is_usable}

    async def _try_to_restore_previous_state(self) -> bool:
        """Try to restore the previous state of the pool from the KVS."""
        if not self._persistence_enabled:
            logger.warning('Persistence is disabled, however, the state restoration was triggered.')

        if not self._kvs:
            logger.warning('SessionPool restoration failed: KVS not initialized. Did you forget to call __aenter__?')
            return False

        previous_state = await self._kvs.get_value(key=self._persist_state_key)

        if previous_state is None:
            logger.debug('SessionPool restoration skipped: No previous state found.')
            return False

        previous_session_pool = SessionPoolModel.model_validate(previous_state)

        for session_model in previous_session_pool.sessions:
            session = Session.from_model(model=session_model)
            if session.is_usable:
                self._sessions[session.id] = session

        return True

    async def _persist_state(self, event_data: EventPersistStateData) -> None:
        """Persist the state of the pool in the KVS."""
        logger.debug(f'Persisting state of the SessionPool (event_data={event_data}).')

        if not self._persistence_enabled:
            logger.warning('Persistence is disabled, however, the state persistence event was triggered.')

        if not self._kvs:
            logger.warning('SessionPool persisting failed: KVS not initialized. Did you forget to call __aenter__?')
            return

        session_pool_state = self.get_state(as_dict=True)
        await self._kvs.set_value(key=self._persist_state_key, value=session_pool_state)
